import torch
import torch.nn as nn
#from torchvision.models.utils import load_state_dict_from_url
from torch.hub import load_state_dict_from_url
model_urls = {
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
}


class SELayer(nn.Module):
    def __init__(self, channel, reduction=16):
        super(SELayer, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(channel, channel // reduction, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(channel // reduction, channel, bias=False),
            nn.Sigmoid()
        )

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y.expand_as(x)


def conv3x3(in_planes, out_planes, stride=1):
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)


class SEBasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None,
                 *, reduction=16):
        super(SEBasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes, 1)
        self.bn2 = nn.BatchNorm2d(planes)
        self.se = SELayer(planes, reduction)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.se(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class SEBottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None,
                 *, reduction=16):
        super(SEBottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.se = SELayer(planes * 4, reduction)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)
        out = self.se(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes, out_planes, stride=1):
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None):
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)

        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)

        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self, block, layers, num_classes=1000, zero_init_residual=False,
                 groups=1, width_per_group=64, replace_stride_with_dilation=None,
                 norm_layer=None):
        super(ResNet, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer

        self.inplanes = 64
        self.dilation = 1
        if replace_stride_with_dilation is None:
            replace_stride_with_dilation = [False, False, False]

        if len(replace_stride_with_dilation) != 3:
            raise ValueError("replace_stride_with_dilation should be None "
                             "or a 3-element tuple, got {}".format(replace_stride_with_dilation))

        self.block = block
        self.groups = groups
        self.base_width = width_per_group

        # 224,224,3 -> 112,112,64
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=7, stride=2, padding=3,
                               bias=False)
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)

        # 112,112,64 -> 56,56,64
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        # 56,56,64 -> 56,56,256
        self.layer1 = self._make_layer(block, 64, layers[0])

        # 56,56,256 -> 28,28,512
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2,
                                       dilate=replace_stride_with_dilation[0])

        # 28,28,512 -> 14,14,1024
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2,
                                       dilate=replace_stride_with_dilation[1])

        # 14,14,1024 -> 7,7,2048
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2,
                                       dilate=replace_stride_with_dilation[2])

        # 7,7,2048 -> 2048
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        # 2048 -> num_classes
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, SEBottleneck):
                    nn.init.constant_(m.bn3.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        # Conv_block
        layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            # identity_block
            layers.append(block(self.inplanes, planes, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        return x

    def freeze_backbone(self):
        backbone = [self.conv1, self.bn1, self.layer1, self.layer2, self.layer3, self.layer4]
        for module in backbone:
            for param in module.parameters():
                param.requires_grad = False

    def Unfreeze_backbone(self):
        backbone = [self.conv1, self.bn1, self.layer1, self.layer2, self.layer3, self.layer4]
        for module in backbone:
            for param in module.parameters():
                param.requires_grad = True


def se_resnet18(num_classes=1000, pretrained=False):
    """Constructs a ResNet-50 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(SEBasicBlock, [2, 2, 2, 2], num_classes=num_classes)
    # model.avgpool = nn.AdaptiveAvgPool2d(1)
    if pretrained:
        pretrained_dict = load_state_dict_from_url(
            "https://download.pytorch.org/models/resnet18-5c106cde.pth")
        model_dict = model.state_dict()
        pretrained_dict = {k: v for k, v in pretrained_dict.items() if (k in model_dict and 'fc' not in k)}
        model_dict.update(pretrained_dict)
        model.load_state_dict(model_dict)
        print("成功加载预训练权重")
        # model.load_state_dict(load_state_dict_from_url(
        #     "https://github.com/moskomule/senet.pytorch/releases/download/archive/seresnet50-60a8950a85b2b.pkl"))
    return model


def se_resnet34(num_classes=1000, pretrained=False):
    """Constructs a ResNet-50 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(SEBasicBlock, [3, 4, 6, 3], num_classes=num_classes)
    # model.avgpool = nn.AdaptiveAvgPool2d(1)
    if pretrained:
        pretrained_dict = load_state_dict_from_url(
            "https://download.pytorch.org/models/resnet34-333f7ec4.pth")
        model_dict = model.state_dict()
        pretrained_dict = {k: v for k, v in pretrained_dict.items() if (k in model_dict and 'fc' not in k)}
        model_dict.update(pretrained_dict)
        model.load_state_dict(model_dict)
        print("成功加载预训练权重")
        # model.load_state_dict(load_state_dict_from_url(
        #     "https://github.com/moskomule/senet.pytorch/releases/download/archive/seresnet50-60a8950a85b2b.pkl"))
    return model


def se_resnet50(num_classes=1000, pretrained=False):
    """Constructs a ResNet-50 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(SEBottleneck, [3, 4, 6, 3], num_classes=num_classes)
    # model.avgpool = nn.AdaptiveAvgPool2d(1)
    if pretrained:
        pretrained_dict = load_state_dict_from_url(
            "https://github.com/moskomule/senet.pytorch/releases/download/archive/seresnet50-60a8950a85b2b.pkl")
        model_dict = model.state_dict()
        pretrained_dict = {k: v for k, v in pretrained_dict.items() if (k in model_dict and 'fc' not in k)}
        model_dict.update(pretrained_dict)
        model.load_state_dict(model_dict)
        print("成功加载预训练权重")
        # model.load_state_dict(load_state_dict_from_url(
        #     "https://github.com/moskomule/senet.pytorch/releases/download/archive/seresnet50-60a8950a85b2b.pkl"))
    return model


def resnet50(pretrained=False, progress=True):
    model = ResNet(SEBottleneck, [3, 4, 6, 3])
    if pretrained:
        state_dict = load_state_dict_from_url(model_urls['resnet50'], model_dir='./model_data',
                                              progress=progress)
        model.load_state_dict(state_dict)

    # if num_classes!=1000:
    #     model.fc = nn.Linear(512 * model.block.expansion, num_classes)
    return model


class SE_Multimodel18_no_text(nn.Module):
    def __init__(self, pretrained, num_classes):
        super(SE_Multimodel18_no_text, self).__init__()
        self.num_classes = num_classes
        self.model = se_resnet18(num_classes=num_classes, pretrained=pretrained)
        self.model_CDIF = se_resnet18(num_classes=num_classes, pretrained=pretrained)
        self.fc = nn.Sequential(
            nn.Linear(1024, 256),  # 4096+92    #4096+59  #1024+59=1083
            nn.ReLU(),
            nn.Dropout(0.3),
            nn.Linear(256, self.num_classes)
        )

    def forward(self, x, x_withCDFI):
        x = self.model(x)
        x = torch.flatten(x, 1)
        x_withCDFI = self.model_CDIF(x_withCDFI)
        x_withCDFI = torch.flatten(x_withCDFI, 1)
        output = torch.cat((x, x_withCDFI), 1)
        output = self.fc(output)
        return output

    def Multi_freeze_backbone(self):
        self.model.freeze_backbone()
        self.model_CDIF.freeze_backbone()

    def Multi_Unfreeze_backbone(self):
        self.model.Unfreeze_backbone()
        self.model_CDIF.Unfreeze_backbone()


class SE_Three_model18_no_text(nn.Module):
    def __init__(self, pretrained, num_classes):
        super(SE_Three_model18_no_text, self).__init__()
        self.num_classes = num_classes
        self.model = se_resnet18(num_classes=num_classes, pretrained=pretrained)
        self.model_mask = se_resnet18(num_classes=num_classes, pretrained=pretrained)
        self.model_CDIF = se_resnet18(num_classes=num_classes, pretrained=pretrained)
        self.fc = nn.Sequential(
            nn.Linear(1536, 256),  # 4096+92    #4096+59  #1024+59=1083
            nn.ReLU(),
            nn.Dropout(0.3),
            nn.Linear(256, self.num_classes)
        )

    def forward(self, x, x_mask, x_withCDFI):
        x = self.model(x)
        x = torch.flatten(x, 1)

        x_mask = self.model_mask(x_mask)
        x_mask = torch.flatten(x_mask, 1)

        x_withCDFI = self.model_CDIF(x_withCDFI)
        x_withCDFI = torch.flatten(x_withCDFI, 1)

        output = torch.cat((x, x_withCDFI), 1)
        output = torch.cat((output, x_mask), 1)
        output = self.fc(output)
        return output

    def Multi_freeze_backbone(self):
        self.model.freeze_backbone()
        self.model_CDIF.freeze_backbone()

    def Multi_Unfreeze_backbone(self):
        self.model.Unfreeze_backbone()
        self.model_CDIF.Unfreeze_backbone()


class SE_Multimodel34_no_text(nn.Module):
    def __init__(self, pretrained, num_classes):
        super(SE_Multimodel34_no_text, self).__init__()
        self.num_classes = num_classes
        self.model = se_resnet34(num_classes=num_classes, pretrained=pretrained)
        self.model_CDIF = se_resnet34(num_classes=num_classes, pretrained=pretrained)
        self.fc = nn.Sequential(
            nn.Linear(1083, 256),  # 512*2+92    #512*2+59
            nn.ReLU(),
            nn.Dropout(0.3),
            nn.Linear(256, self.num_classes)
        )

    def forward(self, x, x_withCDFI, x_info):
        x = self.model(x)
        x = torch.flatten(x, 1)
        x_withCDFI = self.model_CDIF(x_withCDFI)
        x_withCDFI = torch.flatten(x_withCDFI, 1)
        output = torch.cat((x, x_withCDFI), 1)
        output = torch.cat((output, x_info), 1)
        output = self.fc(output)
        return output

    def Multi_freeze_backbone(self):
        self.model.freeze_backbone()
        self.model_CDIF.freeze_backbone()

    def Multi_Unfreeze_backbone(self):
        self.model.Unfreeze_backbone()
        self.model_CDIF.Unfreeze_backbone()


class SE_Multimodel34_no_text(nn.Module):
    def __init__(self, pretrained, num_classes):
        super(SE_Multimodel34_no_text, self).__init__()
        self.num_classes = num_classes
        self.model = se_resnet34(num_classes=num_classes, pretrained=pretrained)
        self.model_CDIF = se_resnet34(num_classes=num_classes, pretrained=pretrained)
        self.fc = nn.Sequential(
            nn.Linear(1024, 256),  # 512*2+92    #512*2+59
            nn.ReLU(),
            nn.Dropout(0.3),
            nn.Linear(256, self.num_classes)
        )

    def forward(self, x, x_withCDFI):  # ,x_info
        x = self.model(x)
        x = torch.flatten(x, 1)
        x_withCDFI = self.model_CDIF(x_withCDFI)
        x_withCDFI = torch.flatten(x_withCDFI, 1)
        output = torch.cat((x, x_withCDFI), 1)
        output = self.fc(output)
        return output

    def Multi_freeze_backbone(self):
        self.model.freeze_backbone()
        self.model_CDIF.freeze_backbone()

    def Multi_Unfreeze_backbone(self):
        self.model.Unfreeze_backbone()
        self.model_CDIF.Unfreeze_backbone()


class SE_Multimodel50_no_text(nn.Module):
    def __init__(self, pretrained, num_classes):
        super(SE_Multimodel50_no_text, self).__init__()
        self.num_classes = num_classes
        self.model = se_resnet50(num_classes=num_classes, pretrained=pretrained)
        self.model_CDIF = se_resnet50(num_classes=num_classes, pretrained=pretrained)
        self.fc = nn.Sequential(
            nn.Linear(4096, 256),  # 4096+92    #4096+59
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(256, self.num_classes)
        )

    def forward(self, x, x_withCDFI):
        x = self.model(x)
        x = torch.flatten(x, 1)
        x_withCDFI = self.model_CDIF(x_withCDFI)
        x_withCDFI = torch.flatten(x_withCDFI, 1)
        output = torch.cat((x, x_withCDFI), 1)
        output = self.fc(output)
        return output

    def Multi_freeze_backbone(self):
        self.model.freeze_backbone()
        self.model_CDIF.freeze_backbone()

    def Multi_Unfreeze_backbone(self):
        self.model.Unfreeze_backbone()
        self.model_CDIF.Unfreeze_backbone()


class SEResnet18_no_text_single(nn.Module):
    def __init__(self, pretrained, num_classes):
        super(SEResnet18_no_text_single, self).__init__()
        self.num_classes = num_classes
        self.model = se_resnet18(num_classes=num_classes, pretrained=pretrained)
        self.fc = nn.Sequential(
            nn.Linear(512, 256),  # 4096+92    #4096+59
            nn.ReLU(),
            nn.Dropout(0.3),
            nn.Linear(256, self.num_classes)
        )

    def forward(self, x):
        x = self.model(x)
        x = torch.flatten(x, 1)
        output = self.fc(x)
        return output

    def Multi_freeze_backbone(self):
        self.model.freeze_backbone()

    def Multi_Unfreeze_backbone(self):
        self.model.Unfreeze_backbone()


if __name__ == "__main__":
    multimodel = SE_Multimodel50_no_text(pretrained=True, num_classes=2)
    input_image = torch.randn(4, 3, 256, 256)
    input_image2 = torch.randn(4, 3, 256, 256)
    output = multimodel(input_image, input_image2)
    print(output.shape)